import os
import json
import logging

from configparser import ConfigParser

# for CLI use
HOME_PATH = os.path.dirname(os.path.abspath(__file__))
ETC_DIR = os.path.join(HOME_PATH, "etc")

DEFAULT_CONFIG = os.path.join(HOME_PATH, ETC_DIR, "config.ini")
GLOBAL_CONFIG_PATH = "/opt/ace/etc/ace_metrics.ini"
USER_CONFIG_PATH = os.path.join(os.path.expanduser("~"), ".config", "ace", "metrics.ini")
ENV_CONFIG_PATH = os.environ["ACE_METRICS_CONFIG_PATH"] if "ACE_METRICS_CONFIG_PATH" in os.environ else None

# Later configs override earlier configs
DEFAULT_CONFIG_PATHS = [DEFAULT_CONFIG, GLOBAL_CONFIG_PATH, USER_CONFIG_PATH]
if ENV_CONFIG_PATH is not None:
    DEFAULT_CONFIG_PATHS.append(ENV_CONFIG_PATH)


def load_config(config_paths: list = DEFAULT_CONFIG_PATHS) -> ConfigParser:
    """Load ACE Metric configuration.

    Args:
      config_paths: List of configuration paths.

    Returns:
      A ConfigParser
    """

    config = ConfigParser()
    finds = []
    for cp in config_paths:
        if cp and os.path.exists(cp):
            logging.debug("Found config file at {}.".format(cp))
            finds.append(cp)
    if not finds:
        logging.error("Didn't find any config files defined at these paths: {}".format(config_paths))
        return False

    config.read(finds)

    return config



"""Alert page for Dash app.

All functions that generate charts will have a @callback decorator along
  with its input and output parameters. They also share the return type
  and an exception as described below. Also, double-clicking on any plot
  will automatically adjust the range of its y-axis. Some charts can
  have 2 plot views: `Default` for regular columns and `Critical`
  for more important columns.

Args:
    relayoutData: Contains layout changes such as hidden traces, or user
      interactions like zooming, resetting axes, etc.
    template_data: Plot template based on chosen theme and color mode.
    current_fig: Existing figure to update, if applicable.

Returns:
    An updated figure or the same figure if no changes are made.

Raises:
    PreventUpdate: If no update is required or possible.
"""

from typing import Dict, List, Union

import dash
import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from dash import Input, Output, State, callback, dcc
from plotly.subplots import make_subplots

from ace_metrics.plotly_dash.app import chart_config
from ace_metrics.plotly_dash.helpers import double_click_reset_y_range, fetch_data, to_fiscal_year, to_ordinal

dash.register_page(__name__, path="/")


def extended_df(df: pd.DataFrame) -> pd.DataFrame:
    """Add more columns to the DataFrame."""
    df = df.copy(deep=True)
    df["total"] = df.drop("month", axis=1).sum(axis=1)
    df["exploitation+installation"] = df["exploitation"] + df["installation"]
    df["recon+weaponization"] = df["reconnaissance"] + df["weaponization"]
    df["actions_on_objectives"] = df["exfil"] + df["damage"]

    return df


def update_side_by_side_chart(
    relayoutData: Dict, current_fig: Dict, template: Dict, table: str, chart_title: str
) -> Union[Dict, go.Figure]:
    """Update or create a side-by-side chart based on provided data.

    Create two subplots with shared x-axis, one for general data and the
      other for business hours data.

    Args:
        table: The name of the table used to generate charts.
        chart_title: Title of the chart.
    """
    if not current_fig and relayoutData:
        df = extended_df(fetch_data([table], True)[table])
        df_bh = extended_df(fetch_data([f"{table}_BH"], True)[f"{table}_BH"])
        included_cols = [
            "month",
            "exploitation+installation",
            "command_and_control",
            "actions_on_objectives",
            "recon+weaponization",
            "false_positive",
            "exploitation",
            "installation",
            "exfil",
            "damage",
            "delivery",
            "policy_violation",
            "grayware",
            "weaponization",
            "reconnaissance",
        ]
        df = df[included_cols]
        df_bh = df_bh[included_cols]

        fig = make_subplots(
            rows=1,
            cols=2,
            shared_xaxes=True,
            subplot_titles=(
                f"{chart_title}",
                f"{chart_title} (Business Hours)",
            ),
            horizontal_spacing=0.037,
        )
        color = px.colors.qualitative.Light24
        for i, col in enumerate(df.columns[1:]):
            fig.add_trace(
                go.Scatter(
                    x=df["month"],
                    y=df[col],
                    mode="lines+markers",
                    name=col,
                    line=dict(color=color[i]),
                    legendgroup=f"group{i}",
                ),
                row=1,
                col=1,
            )
            fig.add_trace(
                go.Scatter(
                    x=df_bh["month"],
                    y=df_bh[col],
                    mode="lines+markers",
                    name=col,
                    line=dict(color=color[i]),
                    legendgroup=f"group{i}",
                    showlegend=False,
                ),
                row=1,
                col=2,
            )
        fig.update_xaxes(range=[len(df.index) - 13, len(df.index) - 1])
        fig.update_yaxes(range=[0, df.tail(13).iloc[:, 1:].max().max()])
        fig.update_yaxes(range=[0, df_bh.tail(13).iloc[:, 1:].max().max()], row=1, col=2)

        number_of_months = len(df["month"])
        y1_min, y1_max, y2_min, y2_max = float("inf"), float("-inf"), float("inf"), float("-inf")
        y1_min_crit, y1_max_crit, y2_min_crit, y2_max_crit = float("inf"), float("-inf"), float("inf"), float("-inf")
        x_axis = fig.data[0]["x"]
        for trace in fig.data:
            x_values, y_values = trace["x"], trace["y"]
            for x, y in zip(x_values, y_values):
                if x_axis[number_of_months - 13] <= x <= x_axis[number_of_months - 1]:
                    if trace.legendgroup not in ["group0", "group2", "group3"]:  # Default View
                        if trace["showlegend"] is None:  # First subplot
                            y1_min = min(y1_min, y)
                            y1_max = max(y1_max, y)
                        else:
                            y2_min = min(y2_min, y)
                            y2_max = max(y2_max, y)
                    else:  # Critical View
                        if trace["showlegend"] is None:  # First subplot
                            y1_min_crit = min(y1_min_crit, y)
                            y1_max_crit = max(y1_max_crit, y)
                        else:
                            y2_min_crit = min(y2_min_crit, y)
                            y2_max_crit = max(y2_max_crit, y)
        buttons = [
            dict(
                method="update",
                label="Default",
                args=[
                    {
                        "visible": [
                            True if trace.legendgroup not in ["group0", "group2", "group3"] else "legendonly"
                            for trace in fig.data
                        ],
                    },
                    {
                        "xaxis.range[0]": number_of_months - 13,
                        "xaxis2.range[0]": number_of_months - 13,
                        "xaxis.range[1]": number_of_months - 1,
                        "xaxis2.range[1]": number_of_months - 1,
                        "yaxis.range[0]": y1_min,
                        "yaxis2.range[0]": y2_min,
                        "yaxis.range[1]": y1_max,
                        "yaxis2.range[1]": y2_max,
                    },
                ],
            ),
            dict(
                method="update",
                label="Critical",
                args=[
                    {
                        "visible": [
                            True if trace.legendgroup in ["group0", "group1", "group2", "group3"] else "legendonly"
                            for trace in fig.data
                        ],
                    },
                    {
                        "xaxis.range[0]": number_of_months - 13,
                        "xaxis2.range[0]": number_of_months - 13,
                        "xaxis.range[1]": number_of_months - 1,
                        "yaxis.range[0]": y1_min_crit,
                        "yaxis2.range[0]": y2_min_crit,
                        "yaxis.range[1]": y1_max_crit,
                        "yaxis2.range[1]": y2_max_crit,
                    },
                ],
            ),
        ]

        fig.update_annotations(yshift=60)
        fig.update_layout(
            updatemenus=[dict(type="buttons", direction="left", x=0.5, xanchor="center", y=1.27, buttons=buttons)],
            legend={
                "title_text": "",
                "orientation": "h",
                "x": 0.5,
                "xanchor": "center",
                "y": 1.18,
                "bgcolor": "rgba(0,0,0,0)",
            },
            margin=dict(t=114, l=0, r=0, b=0),
            template=template,
        )

        fig.update_yaxes(title_text="hours", row=1, col=1)

        return fig

    if current_fig and relayoutData == {}:
        # Update the y-axis range
        current_fig["layout"]["yaxis"]["range"], current_fig["layout"]["yaxis2"]["range"] = double_click_reset_y_range(
            current_fig["data"], current_fig["layout"], True
        )

        return current_fig

    raise dash.exceptions.PreventUpdate


layout = dbc.Container(
    [
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(id="alert-count", config=chart_config, style={"height": 405}),
                    ],
                    width={"size": 12, "order": "last", "offset": 0},
                ),
            ],
            className="mb-3",
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(
                            id="yearly-alert-rank",
                            config=chart_config,
                            style={"height": 405},
                        ),
                    ],
                    width={"size": 6, "order": "last", "offset": 0},
                ),
                dbc.Col(
                    [
                        dcc.Graph(
                            id="alert-sla",
                            config=chart_config,
                            style={"height": 405},
                        ),
                    ],
                    width={"size": 6, "order": "last", "offset": 0},
                ),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(id="alert-by-analyst", config=chart_config, style={"height": 405}),
                    ],
                    width={"size": 6, "order": "last", "offset": 0},
                ),
                dbc.Col(
                    [
                        dcc.Graph(id="alert-count-by-type", config=chart_config, style={"height": 405}),
                    ],
                    width={"size": 6, "order": "last", "offset": 0},
                ),
            ],
            className="mb-3",
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(id="total-open-time", config=chart_config, style={"height": 480}),
                    ],
                    width={"size": 12, "order": "last", "offset": 0},
                ),
            ],
            className="mb-3",
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(id="avg-time-to-dispo", config=chart_config, style={"height": 480}),
                    ],
                    width={"size": 12, "order": "last", "offset": 0},
                ),
            ],
            className="mb-3",
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(id="std-time-to-dispo", config=chart_config, style={"height": 480}),
                    ],
                    width={"size": 12, "order": "last", "offset": 0},
                ),
            ],
            className="mb-3",
        ),
    ]
)


@callback(
    Output("alert-count", "figure"),
    [Input("alert-count", "relayoutData"), Input("theme-template-store", "data")],
    [State("alert-count", "figure")],
)
def update_alert_count(relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict) -> Union[go.Figure, Dict]:
    """Update or create a monthly alert count line chart."""
    if not current_fig and relayoutData:
        df = extended_df(fetch_data(["alert_count"], True)["alert_count"])
        df = df[
            [
                "month",
                "total",
                "exploitation+installation",
                "command_and_control",
                "actions_on_objectives",
                "recon+weaponization",
                "false_positive",
                "exploitation",
                "installation",
                "exfil",
                "damage",
                "delivery",
                "policy_violation",
                "grayware",
                "weaponization",
                "reconnaissance",
                # "reviewed",
                # "unknown",
            ]
        ]

        fig = px.line(
            df,
            x="month",
            y=df.columns[1:],
            labels={"value": "alert count"},
            title="Monthly Alert Counts",
            range_x=[len(df["month"]) - 61, len(df["month"]) - 1],  # 60 months back of data for default view
            range_y=[0, df.tail(61).iloc[:, 1].max().max()],
            template=template_data["template"],
        )

        buttons = [
            dict(
                method="restyle",
                label="Default",
                args=[
                    {
                        "visible": [
                            True
                            if x
                            not in [
                                "exploitation+installation",
                                "command_and_control",
                                "actions_on_objectives",
                                "recon+weaponization",
                            ]
                            else "legendonly"
                            for x in df.columns[1:]
                        ]
                    }
                ],
            ),
            dict(
                method="restyle",
                label="Critical",
                args=[
                    {
                        "visible": [
                            True
                            if x
                            in [
                                "exploitation+installation",
                                "command_and_control",
                                "actions_on_objectives",
                                "recon+weaponization",
                            ]
                            else "legendonly"
                            for x in df.columns[1:]
                        ]
                    }
                ],
            ),
        ]

        fig.update_layout(
            updatemenus=[
                dict(
                    type="buttons",
                    direction="left",
                    x=1.023,
                    xanchor="left",
                    y=-0.05,
                    buttons=buttons,
                )
            ],
            legend={"title_text": ""},
            margin=dict(t=43, l=0),
        )

        for trace in fig.data:
            trace.mode = "lines+markers"
            trace.marker = dict(size=5, color=trace.line.color)

        return fig

    if current_fig and relayoutData == {}:
        current_fig["layout"]["yaxis"]["range"] = double_click_reset_y_range(current_fig["data"], current_fig["layout"])
        return current_fig

    raise dash.exceptions.PreventUpdate


@callback(
    Output("yearly-alert-rank", "figure"),
    [Input("yearly-alert-rank", "relayoutData"), Input("theme-template-store", "data")],
    [State("yearly-alert-rank", "figure")],
)
def update_yearly_alert_ranks(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> Union[Dict, go.Figure]:
    """Update or create bar charts to rank alert types over a period."""

    def _create_plot(selected_columns: List[str]):
        df = extended_df(fetch_data(["alert_count"], True)["alert_count"]).set_index("month")
        last_6_months = df.tail(6).sum()
        last_6_months.name = "last_6_months"
        last_12_months = df.tail(12).sum()
        last_12_months.name = "last_12_months"

        df["fiscal_year"] = df.index.astype(str).to_series().apply(to_fiscal_year)
        df = df.groupby("fiscal_year").sum()
        df = df.append([last_6_months, last_12_months])

        df = df[selected_columns].T
        sorted_df = pd.DataFrame()

        for year_time in df.columns:
            temp_df = df.sort_values(by=year_time, ascending=False).reset_index()
            temp_df["rank"] = temp_df.index.map(lambda x: f"{to_ordinal(x+1)}")
            temp_df = temp_df.rename(columns={"index": f"alert_type_{year_time}"})
            temp_df = temp_df[["rank", f"alert_type_{year_time}", year_time]]

            if sorted_df.empty:
                sorted_df = temp_df
            else:
                sorted_df = pd.merge(sorted_df, temp_df, on="rank", how="outer")
        # Create the figure
        fig = go.Figure()
        for year_time in df.columns:
            fig.add_trace(
                go.Bar(
                    x=sorted_df["rank"],
                    y=sorted_df[year_time],
                    name=f"{year_time}",
                    text=sorted_df[f"alert_type_{year_time}"],
                    hoverinfo="y+text",
                )
            )

        buttons = [
            dict(method="relayout", label=option, args=[{"xaxis.title": option}]) for option in ["Default", "Critical"]
        ]
        fig.update_layout(
            title="Yearly Alert Disposition Rankings (Excl. FPs)",
            xaxis_title="rank",
            yaxis_title="alert count",
            barmode="group",
            xaxis={"range": [-0.5, 4.5]} if "reconnaissance" in selected_columns else {"range": [-0.5, 3.5]},
            updatemenus=[
                dict(
                    active=0 if "reconnaissance" in selected_columns else 1,
                    direction="left",
                    type="buttons",
                    x=0.75,
                    xanchor="left",
                    y=-0.07,
                    buttons=buttons,
                )
            ],
            legend={
                "x": 1,
                "xanchor": "right",
                "bgcolor": "rgba(0,0,0,0)",
            },
            margin=dict(t=55, l=0, r=0, b=55),
            template=template_data["template"],
        )
        for i, trace in enumerate(fig.data):
            if trace.name in ["last_12_months", df.columns[-4]]:
                trace.visible = True
            else:
                trace.visible = "legendonly"

        return fig

    critical_columns = ["exploitation+installation", "recon+weaponization", "actions_on_objectives", "delivery"]
    default_columns = [
        col
        for col in extended_df(fetch_data(["alert_count"], True)["alert_count"]).columns.drop(
            ["total", "month", "false_positive"]
        )
        if col not in critical_columns
    ]
    selected_columns = default_columns

    if current_fig and relayoutData and "xaxis.title" in relayoutData:
        plot_type = relayoutData["xaxis.title"]
        if plot_type == "Default":
            selected_columns = default_columns
        else:
            selected_columns = critical_columns
        return _create_plot(selected_columns)

    if not current_fig and relayoutData:
        return _create_plot(selected_columns)

    if current_fig and relayoutData == {}:
        current_fig["layout"]["yaxis"]["autorange"] = True
        current_fig["layout"]["xaxis"]["range"] = (
            [-0.5, 4.5] if "reconnaissance" in current_fig["data"][0]["text"] else [-0.5, 3.5]
        )
        return current_fig
    raise dash.exceptions.PreventUpdate


@callback(
    Output("alert-sla", "figure"),
    [Input("alert-sla", "relayoutData"), Input("theme-template-store", "data")],
    [State("alert-sla", "figure")],
)
def update_alert_sla(relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict) -> Union[go.Figure, Dict]:
    """Update or create a monthly alerts SLA line chart."""
    if not current_fig and relayoutData:
        df = fetch_data(["overall_operating_alert"], True)["overall_operating_alert"].set_index("month")
        df = df[["bh_cycletime_avg", "real_cycletime_avg"]]
        fig = px.line(
            df,
            x=df.index,
            y=df.columns,
            labels={"value": ""},
            title="Monthly Alerts SLA",
            range_x=[len(df.index) - 25, len(df.index) - 1],
            range_y=[0, df.tail(25).max().max()],
        )

        fig.update_layout(
            legend=dict(
                x=0,
                y=1.1,
                traceorder="normal",
                orientation="h",
                title=dict(text=""),
                bgcolor="rgba(0,0,0,0)",
            ),
            margin=dict(t=55, l=0, r=0, b=0),
            template=template_data["template"],
        )

        for trace in fig.data:
            if trace.name == "bh_cycletime_avg":
                trace.name = "Average Business Cycle-Time (SLA)"
            else:
                trace.name = "Average Raw Cycle-Time"
            trace.mode = "lines+markers"
            trace.marker = dict(size=5, color=trace.line.color)

        return fig

    if current_fig and relayoutData == {}:
        current_fig["layout"]["yaxis"]["range"] = double_click_reset_y_range(current_fig["data"], current_fig["layout"])
        return current_fig

    raise dash.exceptions.PreventUpdate


@callback(
    Output("alert-by-analyst", "figure"),
    [Input("alert-by-analyst", "relayoutData"), Input("theme-template-store", "data")],
    [State("alert-by-analyst", "figure")],
)
def update_alert_by_analyst(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> Union[Dict, go.Figure]:
    """Update or create a line chart for alerts dispo'ed by analyst."""
    if not current_fig and relayoutData:
        df = fetch_data(["analyst_alert_quantities"], True)["analyst_alert_quantities"].set_index("month")
        df.sort_index(inplace=True)

        # Drop inactive users since last year
        sum_last_12_months = df.tail(12).sum()
        df.drop(columns=sum_last_12_months[sum_last_12_months == 0].index.tolist(), inplace=True)

        fig = px.line(
            df,
            x=df.index,
            y=df.columns,
            labels={"value": "alert count"},
            title="Alert Quantities by Analyst",
            range_x=[len(df.index) - 13, len(df.index) - 1],  # 12 months back of data for default view
            range_y=[0, df.tail(13).max().max()],
        )

        fig.update_layout(
            legend={
                "title_text": "",
                "orientation": "h",
                "y": 1.18,
                "bgcolor": "rgba(0,0,0,0)",
            },
            margin=dict(t=100, l=0, r=0, b=0),
            template=template_data["template"],
        )
        return fig

    if current_fig and relayoutData == {}:
        current_fig["layout"]["yaxis"]["range"] = double_click_reset_y_range(current_fig["data"], current_fig["layout"])
        return current_fig

    raise dash.exceptions.PreventUpdate


@callback(
    Output("alert-count-by-type", "figure"),
    [Input("alert-count-by-type", "relayoutData"), Input("theme-template-store", "data")],
    [State("alert-count-by-type", "figure")],
)
def update_alert_count_by_type(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> Union[go.Figure, Dict]:
    """Update or create a line chart for alert quantities by type."""
    if not current_fig and relayoutData:
        df = fetch_data(["alert_type_quantities"], True)["alert_type_quantities"].set_index("month")
        df.sort_index(inplace=True)

        fig = px.line(
            df,
            x=df.index,
            y=df.columns,
            labels={"value": "alert count"},
            title="Alert Type Category Quantities",
            range_x=[len(df.index) - 13, len(df.index) - 1],  # 12 months back of data for default view
            range_y=[0, df.tail(13).max().max()],
            template=template_data["template"],
        )

        fig.update_layout(
            legend={
                "title_text": "",
                "orientation": "h",
                "y": 1.18,
                "bgcolor": "rgba(0,0,0,0)",
            },
            margin=dict(t=100, l=0, r=0, b=0),
        )
        return fig

    if current_fig and relayoutData == {}:
        current_fig["layout"]["yaxis"]["range"] = double_click_reset_y_range(current_fig["data"], current_fig["layout"])
        return current_fig

    raise dash.exceptions.PreventUpdate


@callback(
    Output("total-open-time", "figure"),
    [Input("total-open-time", "relayoutData"), Input("theme-template-store", "data")],
    [State("total-open-time", "figure")],
)
def update_total_open_time(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> Union[Dict, go.Figure]:
    """Generate Total Open Time chart."""
    return update_side_by_side_chart(
        relayoutData, current_fig, template_data["template"], "cycle_time_sum", "Total Open Time"
    )


@callback(
    Output("avg-time-to-dispo", "figure"),
    [Input("avg-time-to-dispo", "relayoutData"), Input("theme-template-store", "data")],
    [State("avg-time-to-dispo", "figure")],
)
def update_avg_time_to_dispo(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> Union[Dict, go.Figure]:
    """Generate Average Time to Disposition chart."""
    return update_side_by_side_chart(
        relayoutData, current_fig, template_data["template"], "cycle_time_mean", "Average Time to Disposition"
    )


@callback(
    Output("std-time-to-dispo", "figure"),
    [Input("std-time-to-dispo", "relayoutData"), Input("theme-template-store", "data")],
    [State("std-time-to-dispo", "figure")],
)
def update_std_time_to_dispo(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> Union[Dict, go.Figure]:
    """Generate Standard Deviation for Time to Disposition chart."""
    return update_side_by_side_chart(
        relayoutData,
        current_fig,
        template_data["template"],
        "cycle_time_std",
        "Standard Deviation for Time to Disposition",
    )



"""Operations for local SQLite database to store all metrics."""

import logging
import os
import sqlite3
import sys
import warnings
from datetime import datetime, timedelta
from typing import Any, Dict

import sqlite_utils
from dateutil.relativedelta import relativedelta

from ace_metrics.alerts import (
    VALID_ALERT_STATS,
    define_business_time,
    generate_hours_of_operation_summary_table,
    generate_overall_summary_table,
    get_alerts_between_dates,
    statistics_by_month_by_dispo,
)

from ..alerts.alert_types import alert_type_quantities_by_category_by_month
from ..alerts.users import alert_quantities_by_user_by_month
from ..config import DEFAULT_CONFIG_PATHS, load_config
from ..helpers import connect_to_database
from .data_exceptions import apply_alert_data_exceptions

logging.basicConfig(level=logging.INFO)
warnings.filterwarnings(action="ignore", category=FutureWarning)
warnings.filterwarnings("ignore", message=".*pandas only supports SQLAlchemy connectable.*")

DEFAULT_DATA_DIR = os.path.join("/", "opt", "ace", "data", "stats", "metrics")
config = load_config(DEFAULT_CONFIG_PATHS)
if not config:
    logging.error("You must define and supply a configuration. See `ace-metrics -h | grep config`")
    sys.exit(1)
try:
    DATABASE_CONFIG = config["database_default"]
    DASH_CONFIG = config["plotly_dash"]
except KeyError:
    logging.error("database_default and plotly_dash sections need to be defined in the configuration file.")
    sys.exit(1)

if not DASH_CONFIG.get("enabled", None) or not DASH_CONFIG.getboolean("enabled"):
    logging.warning("Dash is not enabled. Exiting...")
    sys.exit(1)

DATA_DIR = DATABASE_CONFIG.get("data_dir", DEFAULT_DATA_DIR) or DEFAULT_DATA_DIR


def floor_datetime_month(dt):
    """Floor a datetime to the absolute beginning of the month."""
    return dt - timedelta(days=dt.day - 1, hours=dt.hour, minutes=dt.minute, seconds=dt.second)


def get_alert_tables_by_month(months_ago: int) -> Dict[str, Any]:
    """Retrieve alert tables grouped by month.

    Args:
        months_ago: Number of months before the current date to
          calculate the start date for alert data retrieval.

    Returns:
        A dictionary mapping table names to their respective alert data
          aggregated by month.
    """
    end_date = datetime.now()
    start_date = floor_datetime_month(end_date - relativedelta(months=months_ago))
    db = connect_to_database(DATABASE_CONFIG)
    business_hours = define_business_time()
    logging.info(f"Querying alerts data from {start_date} up until {end_date} (now).")

    alerts = get_alerts_between_dates(start_date, end_date, db, selected_companies=["ashland"])
    exception_list = DATABASE_CONFIG["alert_data_exception_list"].split(",")
    alerts = apply_alert_data_exceptions(config, alerts, exception_list)
    alert_stat_map = statistics_by_month_by_dispo(alerts)
    alert_stat_map_bh = statistics_by_month_by_dispo(alerts, business_hours=business_hours)
    for stat in VALID_ALERT_STATS:
        if "time" in stat:
            alert_stat_map_bh[f"{stat}_BH"] = alert_stat_map_bh.pop(f"{stat}")
    database_table_map = alert_stat_map | alert_stat_map_bh

    database_table_map["analyst_alert_quantities"] = alert_quantities_by_user_by_month(start_date, end_date, db)

    alert_type_categories_key = {}
    for k, v in config["alert_type_categories_key"].items():
        alert_type_categories_key[k] = v.split(",")
    database_table_map["alert_type_quantities"] = alert_type_quantities_by_category_by_month(
        start_date, end_date, db, alert_type_categories_key
    )

    database_table_map["hours_of_operation"] = generate_hours_of_operation_summary_table(
        alerts.copy(), business_hours, True
    )
    database_table_map["overall_operating_alert"] = generate_overall_summary_table(alerts.copy(), business_hours, True)
    return database_table_map


def initialize_database() -> bool:
    """Initialize or reuse a local SQLite database."""
    database_path = f"{DATA_DIR}/ace_metrics_database.sqlite"
    if os.path.isfile(database_path):
        logging.info(f"A database already exists at {database_path}. Proceeding using this database...")
        return True
    logging.info(f"Database does not exist. Creating a new one at {database_path}...")
    database_table_map = get_alert_tables_by_month(DATABASE_CONFIG.getint("alert_data_scope_months_ago", fallback=60))

    with sqlite3.connect(f"{DATA_DIR}/ace_metrics_database.sqlite") as conn:
        for table_name, df in database_table_map.items():
            logging.info(f"Inserting {table_name} table into sqlite database...")
            df.to_sql(table_name, conn, if_exists="replace", index=True, dtype={"month": "TEXT PRIMARY KEY"})

    # Make a copy of the initialized database
    with open(f"{DATA_DIR}/ace_metrics_database.sqlite", "rb") as f1:
        with open(f"{DATA_DIR}/ace_metrics_database_init.sqlite", "wb") as f2:
            f2.write(f1.read())
    return True


def update_database() -> None:
    """Upsert the local database with alert data from the past month.

    Returns:
        A timestamp string when the update was completed.
    """
    logging.info("Updating local metrics database...")
    update_table_map = get_alert_tables_by_month(1)

    with sqlite3.connect(f"{DATA_DIR}/ace_metrics_database.sqlite") as conn:
        db = sqlite_utils.Database(conn)
        for table in update_table_map:
            db[table].upsert_all(update_table_map[table].reset_index().to_dict(orient="records"), pk="month")

    # Update modifification time
    os.utime(f"{DATA_DIR}/ace_metrics_database.sqlite", None)
    logging.info("Database update completed.")


def get_last_update_time() -> str:
    """Get last modified time of database."""
    timestamp = os.path.getmtime(f"{DATA_DIR}/ace_metrics_database.sqlite")
    return datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S")



"""Metric functions for working with users and alerts.
"""

import os
import logging

import pymysql
import businesstime
import pandas as pd

from typing import Dict, Mapping
from datetime import datetime

from ..helpers import get_month_keys_between_two_dates
from . import statistics_by_month_by_dispo, VALID_ALERT_STATS, FRIENDLY_STAT_NAME_MAP, ALERTS_BY_MONTH_AND_USER_QUERY

UserMap = Mapping[int, Dict[str, str]]
UserStatMap = Mapping[str, Mapping[str, pd.DataFrame]]


def get_all_users(con: pymysql.connections.Connection) -> UserMap:
    """Get all ACE users.

    Args:
        con: a pymysql database connectable

    Returns:
        A dictionary where the key is the user ID integer and the
        value is a dictionary like:
            {'username': username,
             'display_name': display_name,
             'queue': queue
             }
    """
    cursor = con.cursor()
    cursor.execute("SELECT id,username,display_name,queue,enabled FROM users")
    users = {}
    for user_id, username, display_name, queue, enabled in cursor.fetchall():
        users[user_id] = {"username": username, "display_name": display_name, "queue": queue, "enabled": enabled}
    return users


def generate_user_alert_stats(alerts: pd.DataFrame, users: UserMap, business_hours=False) -> UserStatMap:
    """Generate alert statistics for all users.

    Given a dataframe of alerts, categorize the alerts by the user that
    dispositioned the alerts, and generated alert statistics for each user.

    Args:
        alert: A pd.DataFrame of alerts
        users: A dictionary mapping user IDs to user attributes
        business_hours: A boolean that if True, will calulate time base
          statistics with business hours applied.

    Returns:
        A dictionary where the keys are usernames and the values are alert
        statatistic maps for users.
    """

    all_user_alert_stats = {}
    for user_id in users.keys():
        username = users[user_id]["username"]
        display_name = users[user_id].get("display_name", None)
        user_alerts = alerts[alerts.disposition_user_id == user_id]
        user_alert_stats = statistics_by_month_by_dispo(user_alerts, business_hours=business_hours)
        for stat in VALID_ALERT_STATS:
            if display_name:
                user_alert_stats[stat].name = f"{display_name}: {FRIENDLY_STAT_NAME_MAP[stat]}"
            else:
                user_alert_stats[stat].name = f"{username}: {FRIENDLY_STAT_NAME_MAP[stat]}"
        all_user_alert_stats[user_id] = user_alert_stats

    return all_user_alert_stats


def alert_quantities_by_user_by_month(
    start_date: datetime,
    end_date: datetime,
    con: pymysql.connections.Connection,
    query: str = ALERTS_BY_MONTH_AND_USER_QUERY,
    exclude_analysts_without_data=True,
) -> pd.DataFrame:
    """Get Alert quantities by user and month."""

    months = get_month_keys_between_two_dates(start_date, end_date)
    users = get_all_users(con)

    cursor = con.cursor()

    data = {}
    for uid, udata in users.items():
        month_data = {}
        for month in months:
            cursor.execute(query, (month, udata["username"]))
            stats = cursor.fetchone()
            if stats:
                month_data[month] = stats[0]
        if month_data and exclude_analysts_without_data:
            # don't record users that have no data
            data[udata["username"]] = month_data

    user_dispositions_per_month = pd.DataFrame(data=data)
    user_dispositions_per_month.name = "Alert Quantities by Analyst"
    user_dispositions_per_month.fillna(value=0, inplace=True)
    user_dispositions_per_month.index.name = "month"
    return user_dispositions_per_month



"""Entrypoint for Dash app."""

import copy
import threading
import time
from typing import List

import dash
import dash_bootstrap_components as dbc
import plotly.express as px
import plotly.io as pio
from dash import Dash, Input, Output, clientside_callback, dcc, html
from dash_bootstrap_templates import ThemeChangerAIO
from flask_login import login_required
from flask import Flask
from .database import DASH_CONFIG, get_last_update_time, initialize_database, update_database
from .helpers import change_color_brightness


class FilteredThemeChangerAIO(ThemeChangerAIO):
    """Subclass of ThemeChangerAIO to exclude specific themes."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize ThemeChangerAIO with excluded themes."""
        super().__init__(*args, **kwargs)

        offcanvas = next(child for child in self.children if "Offcanvas" in str(type(child)))
        radio_items = next(child for child in offcanvas.children if "RadioItems" in str(type(child)))

        # Filter out ugly themes
        not_allowed_themes = ["MORPH", "QUARTZ", "SLATE", "CYBORG"]
        radio_items.options = [option for option in radio_items.options if option["label"] not in not_allowed_themes]


filtered_theme_changer = FilteredThemeChangerAIO(
    aio_id="theme-changer",
    button_props={
        "color": "danger",
        "children": "Change theme",
    },
    radio_props={"persistence": True},
)

chart_config = {
    "scrollZoom": True,
    "displaylogo": False,
    "modeBarButtonsToAdd": ["zoom2d", "pan2d", "lasso2d", "toggleSpikelines"],
    "modeBarButtonsToRemove": ["select2d", "resetScale2d"],
    "doubleClick": False,
}


def threaded_db_update() -> None:
    """Run a thread indefinitely that updates the local database.

    Fetch tables from ACE database and update the global DataFrame
      `df_store` and the `last_update_time` variable every 5 minutes.
    """
    while True:
        update_database()
        time.sleep(300)


def generate_main_content() -> dash.page_container:
    """Generate the main content for the Dash application.

    Returns:
        The container for Dash application pages.
    """
    return dash.page_container


def serve_layout() -> dbc.Container:
    """Create the layout of the webapp using Dash Bootstrap Components.

    Define the sidebar and main content area, with the sidebar including
      navigation links, a graph hotkeys accordion, and a theme switcher.

    Returns:
        A Dash Bootstrap container for the full layout of the webapp.
    """
    sidebar = dbc.Col(
        [
            html.H2("ACE Metrics Dashboard", className="display-7"),
            html.Hr(),
            dbc.Nav(
                [
                    dbc.Card(
                        dbc.CardBody(
                            [
                                dbc.NavLink("Alerts", href="/", active="exact", id="alerts-nav"),
                                dbc.NavLink(
                                    "Hours of Operations", href="/hours-of-operations", id="hours-of-operations-nav"
                                ),
                                dbc.NavLink("Events", href="/events", id="events-nav"),
                            ]
                        )
                    ),
                    dbc.Accordion(
                        [
                            dbc.AccordionItem(
                                [
                                    html.P("Z: Pan/Zoom"),
                                    html.P("X: Autoscale X and Y"),
                                    html.P("Double-click: Autoscale Y"),
                                ],
                                title="Graph Hotkeys",
                            ),
                        ],
                        start_collapsed=True,
                    ),
                    html.Span(
                        [
                            html.Div(
                                [
                                    dbc.Label(className="fa fa-moon", html_for="switch"),
                                    dbc.Switch(
                                        id="color-mode-switch",
                                        value=True,
                                        className="d-inline-block ms-1",
                                        persistence=True,
                                    ),
                                    dbc.Label(className="fa fa-sun", html_for="switch"),
                                ],
                                style={"marginBottom": "8px"},
                            ),
                            filtered_theme_changer,
                            dcc.Store(id="theme-template-store"),
                            html.Div(
                                [
                                    html.P(
                                        f"Last updated: {get_last_update_time()}",
                                        style={"marginTop": "15px"},
                                    ),
                                ]
                            ),
                        ],
                        className="d-flex flex-column px-3 pt-4",  # Stack vertically and center align
                    ),
                ],
                vertical=True,
                pills=True,
            ),
        ],
        width={"size": 2, "order": "first", "offset": 0},
        style={
            "position": "sticky",
            "top": 0,
            "height": "100vh",
            "overflowY": "auto",
        },
        className="pt-2",
    )
    return dbc.Container(
        [
            dbc.Row(
                [
                    sidebar,
                    dbc.Col(
                        html.Div([generate_main_content()], id="main-content-div"),
                        style={"maxWidth": "none"},
                        width={"size": 10, "order": "last", "offset": 0},
                    ),
                ],
                className="pt-5",
            ),
            dcc.Store(id="css-store"),
            html.Script(id="css-script"),
        ],
        fluid=True,
    )


def create_dash_app(flask_server: Flask = None) -> Dash:
    """Create and return dash app object."""
    initialize_database()
    if flask_server:
        dash_app = Dash(
            server=flask_server,
            name="Dashboard",
            pages_folder=DASH_CONFIG["pages_folder_dir"],
            assets_folder=DASH_CONFIG["assets_folder_dir"],
            requests_pathname_prefix=DASH_CONFIG["requests_pathname_prefix"],
            routes_pathname_prefix=DASH_CONFIG["routes_pathname_prefix"],
            use_pages=True,
            external_stylesheets=[dbc.icons.FONT_AWESOME],
            suppress_callback_exceptions=True,
        )

        # Protected routes
        for view_function in dash_app.server.view_functions:
            if view_function.startswith(dash_app.config.routes_pathname_prefix):
                dash_app.server.view_functions[view_function] = login_required(
                    dash_app.server.view_functions[view_function]
                )
    else:
        dash_app = Dash(
            __name__,
            use_pages=True,
            external_stylesheets=[dbc.icons.FONT_AWESOME],
            suppress_callback_exceptions=False,
        )
    dash_app.layout = serve_layout

    # Callback to update main content based on theme
    @dash_app.callback(
        [Output("main-content-div", "children"), Output("css-store", "data"), Output("theme-template-store", "data")],
        [
            [Input(filtered_theme_changer.ids.radio("theme-changer"), "value")],
            [Input("color-mode-switch", "value")],
        ],
    )
    def update_main_content_with_theme(selected_theme: List[str], color_mode: List[bool]) -> tuple:
        """Update plot template and CSS for chosen theme and color mode.

        There should be only one element in both parameters.

        Args:
            selected_theme: The theme to apply.
            color_mode: True for light mode, False for dark mode.

        Returns:
            The updated main content and CSS data as a tuple.
        """
        # Extract the theme name from the selected theme path
        theme_name = selected_theme[0].split("/")[-2]
        theme_name = "bootstrap" if theme_name == "css" else theme_name
        # Append '_dark' to the theme name if the color mode is dark
        if not color_mode[0]:
            theme_name += "_dark"
        # Customize the plotly template based on the selected theme
        template_copy = copy.deepcopy(pio.templates[theme_name])
        darker_background = change_color_brightness(
            template_copy["layout"]["plot_bgcolor"], 0.85
        )  # For better contrast
        template_copy["layout"]["plot_bgcolor"] = darker_background
        template_copy["layout"]["colorway"] = (
            px.colors.qualitative.Light24 if "dark" in theme_name else px.colors.qualitative.Dark24
        )

        css_string = f""".Select-menu-outer {{ background-color: {darker_background} !important; }}"""
        return (generate_main_content(), {"css": css_string}, {"template": template_copy})

    # Clientside callback to update theme color scheme
    clientside_callback(
        """
        (switchOn) => {
        switchOn
            ? document.documentElement.setAttribute('data-bs-theme', 'light')
            : document.documentElement.setAttribute('data-bs-theme', 'dark')
        return window.dash_clientside.no_update
        }
        """,
        Output("color-mode-switch", "id"),
        Input("color-mode-switch", "value"),
    )
    # Clientside callback to add custom css
    clientside_callback(
        """
        function(cssData) {
            if (cssData) {
                var styleTag = document.getElementById('custom-style');
                if (!styleTag) {
                    styleTag = document.head.appendChild(document.createElement('style'));
                    styleTag.id = 'custom-style';
                }
                styleTag.innerHTML = cssData.css;
                return;
            }
        }
        """,
        Output("css-script", "children"),
        Input("css-store", "data"),
    )
    return dash_app


if __name__ == "__main__":
    threading.Thread(target=threaded_db_update, daemon=True).start()
    create_dash_app().run(debug=False, port=5432, threaded=False)



"""Allows for configurable custom exceptions."""

import configparser
import importlib
import logging
import re
from typing import Tuple, Union

import pandas as pd

_MODULE_PATH_REGEX = re.compile(r"^([^:]+):([^:]+)(?::(.+))?$")


def SPLIT_MODULE_PATH(module_path) -> Union[bool, Tuple[str, str, str]]:
    """Return a (module, class, instance) tuple from a module path."""
    m = _MODULE_PATH_REGEX.match(module_path)
    if m is None:
        return False

    return m.groups()


def apply_alert_data_exceptions(
    config: configparser.ConfigParser, alerts: pd.DataFrame, exception_list: list
) -> pd.DataFrame:
    """Apply configured exceptions to alerts."""
    for exception_key in exception_list:
        exception_function = None
        _module, _func, _instance = SPLIT_MODULE_PATH(config["alert_data_exceptions"].get(exception_key))
        try:
            m = importlib.import_module(_module)
        except Exception as e:
            logging.error(f"unable to import module {_module}: {e}")
            continue
        try:
            exception_function = getattr(m, _func)
        except Exception as e:
            logging.error(f"unable to import function {_func} from module {_module}: {e}")
            continue
        if exception_function is not None:
            alerts = exception_function(alerts)

    return alerts



"""Metric functions for working with alerts by alert type.
"""

import os
import logging

import pymysql
import businesstime
import pandas as pd

from typing import Dict, Mapping
from datetime import datetime

# from ..helpers import get_companies, apply_company_selection_to_query

from ..helpers import get_month_keys_between_two_dates

from . import (
    ALERTS_BY_ALERT_TYPE_QUERY,
    statistics_by_month_by_dispo,
    VALID_ALERT_STATS,
    FRIENDLY_STAT_NAME_MAP,
    ALERTS_BY_MONTH_AND_TYPE_QUERY,
)

Stat = str
Alerts = pd.DataFrame
AlertType = str
AlertTypeMap = Mapping[AlertType, Alerts]
AlertTypeStatMap = Mapping[AlertType, Mapping[Stat, Alerts]]


def all_alert_types(con: pymysql.connections.Connection) -> list:
    """Get the list of unique alert types.

    Args:
        con: a pymysql database connectable

    Returns:
        The list of unique alert types.
    """

    cursor = con.cursor()
    cursor.execute("select DISTINCT alert_type from alerts")

    alert_types = [_r[0] for _r in cursor.fetchall()]
    return alert_types


def unique_alert_types_between_dates(
    start_date: datetime, end_date: datetime, con: pymysql.connections.Connection
) -> list:
    """Get a list of unique alert types between two dates.

    Args:
        start_date: Get events created on or after this datetime.
        end_date: Get events created on or before this datetime.
        con: a pymysql database connectable

    Returns:
        A list of unique alert types.
    """

    params = [start_date.strftime("%Y-%m-%d %H:%M:%S"), end_date.strftime("%Y-%m-%d %H:%M:%S")]

    cursor = con.cursor()
    cursor.execute("select DISTINCT alert_type from alerts WHERE insert_date BETWEEN %s AND %s", params)

    alert_types = [_r[0] for _r in cursor.fetchall()]
    return alert_types


# TODO: implement company selection here
def count_quantites_by_alert_type(
    start_date: datetime, end_date: datetime, con: pymysql.connections.Connection
) -> pd.DataFrame:
    """Count all alerts by type between two dates.

    Args:
        start_date: Get events created on or after this datetime.
        end_date: Get events created on or before this datetime.
        con: a pymysql database connectable.

    Returns:
        A pd.DataFrame breakdown of the alert_type counts.
    """

    count_query = """SELECT COUNT(alert_type)
                     FROM alerts WHERE alert_type=%s
                     AND insert_date BETWEEN %s AND %s
                  """

    alert_types = unique_alert_types_between_dates(start_date, end_date, con)

    alert_type_counts = {}
    for alert_type in alert_types:
        params = [alert_type, start_date.strftime("%Y-%m-%d %H:%M:%S"), end_date.strftime("%Y-%m-%d %H:%M:%S")]
        cursor = con.cursor()
        cursor.execute(count_query, params)
        results = cursor.fetchone()
        alert_type_counts[alert_type] = results[0]

    if not alert_type_counts:
        empty_df = pd.DataFrame()
        empty_df.name = "Total Alert Type Quantities"
        return empty_df

    at_counts_df = pd.DataFrame.from_dict(alert_type_counts, orient="index")
    at_counts_df.columns = ["Count"]
    at_counts_df.index.name = "Alert Type"
    at_counts_df.name = "Total Alert Type Quantities"
    return at_counts_df


def get_alerts_between_dates_by_type(
    start_date: datetime,
    end_date: datetime,
    con: pymysql.connections.Connection,
    alert_type_query: str = ALERTS_BY_ALERT_TYPE_QUERY,
    selected_companies: list = [],
) -> AlertTypeMap:
    """Get all ACE alerts between two dates, by alert type.

    Query the ACE database for all ACE alerts betweem two dates and
    organize the alerts by alert type.
    If there are selected_companies, only return alerts associated to those companies,
    else all alerts are selected.

    Args:
        start_date: Get alerts created on or after this datetime.
        end_date: Get alert created on or before this datetime.
        con: a pymysql database connectable
        alert_type_query: The str SQL database query to get alerts with.
        selected_companies: A list of companies to select alerts for, by name.
          If the list is empty, all alerts are selected.

    Returns:
        A dictionary where the keys are alert_types and the values are
        pd.DataFrames of those alerts.
    """

    # apply company selection by name
    company_ids = []
    if selected_companies:
        cursor = con.cursor()
        cursor.execute("select * from company")
        for c_id, c_name in cursor.fetchall():
            if c_name in selected_companies:
                company_ids.append(c_id)

    alert_type_query = alert_type_query.format(
        " AND " if company_ids else "",
        "( " + " OR ".join(["company.name=%s" for company in selected_companies]) + ") " if company_ids else "",
    )

    alert_types = unique_alert_types_between_dates(start_date, end_date, con)

    alert_type_dfs = {}
    for alert_type in alert_types:
        params = [alert_type, start_date.strftime("%Y-%m-%d %H:%M:%S"), end_date.strftime("%Y-%m-%d %H:%M:%S")]
        alert_type_df = pd.read_sql_query(alert_type_query, con, params=params)

        alert_type_df.set_index("month", inplace=True)
        alert_type_df.name = alert_type
        alert_type_dfs[alert_type] = alert_type_df

    return alert_type_dfs


def generate_alert_type_stats(alert_type_map: AlertTypeMap, business_hours=False) -> AlertTypeStatMap:
    """Generate alert statistics for all alert types.

    Given a mapping of alert types to a dataframe of alerts,
    generate alert statistics for each alert type.

    Args:
        alert_type_map: A dictionary map of alert_type ->  pd.DataFrame
          of alert_type alerts
        business_hours: A boolean that if True, will calulate time base
          statistics with business hours applied.

    Returns:
        A dictionary where the keys are alert_types and the values are alert
        statatistic maps for the respective alert_types.
    """

    stats_by_alert_type = {}
    for alert_type, at_df in alert_type_map.items():
        alert_type_stats = statistics_by_month_by_dispo(at_df, business_hours=business_hours)
        for stat in VALID_ALERT_STATS:
            alert_type_stats[stat].name = f"{alert_type}: {FRIENDLY_STAT_NAME_MAP[stat]}"
        stats_by_alert_type[alert_type] = alert_type_stats

    return stats_by_alert_type


def alert_type_quantities_by_category_by_month(
    start_date: datetime,
    end_date: datetime,
    con: pymysql.connections.Connection,
    alert_type_categories_key: dict,
    query: str = ALERTS_BY_MONTH_AND_TYPE_QUERY,
    exclude_analysts_without_data=True,
) -> pd.DataFrame:
    """Get Alert quantities by alert type category and month.

    This is useful if you have a diverse set of alert types
    and alerting tools that you want to group into buckets.
    """

    months = get_month_keys_between_two_dates(start_date, end_date)
    alert_types = unique_alert_types_between_dates(start_date, end_date, con)

    # create alert type categories map
    raw_at_accounted_for = []
    alert_type_categories_map = {_k: [] for _k in alert_type_categories_key.keys()}
    for category, at_identifers in alert_type_categories_key.items():
        for at in alert_types:
            if at in raw_at_accounted_for:
                continue
            for _id in at_identifers:
                if at.startswith(_id):
                    alert_type_categories_map[category].append(at)
                    raw_at_accounted_for.append(at)

    def _diff(li1, li2):
        return list(list(set(li1) - set(li2)) + list(set(li2) - set(li1)))

    uncategorized_alert_types = _diff(raw_at_accounted_for, alert_types)
    if uncategorized_alert_types:
        alert_type_categories_map["other"] = uncategorized_alert_types

    cursor = con.cursor()

    data = {}
    for category, atypes in alert_type_categories_map.items():
        month_data = {}
        for month in months:
            month_data[month] = 0
            if not atypes:
                continue
            _query = query.format(" OR ".join(["alert_type=%s" for at in atypes]))
            params = [month] + atypes
            cursor.execute(_query, params)
            stats = cursor.fetchone()
            if stats:
                month_data[month] = stats[0]
        data[category] = month_data

    alert_categories_per_month = pd.DataFrame(data=data)
    alert_categories_per_month.name = "Alert Type Category Quantities"
    alert_categories_per_month.fillna(value=0, inplace=True)
    alert_categories_per_month.index.name = "month"
    for col in alert_categories_per_month.select_dtypes(include=["float64"]):
        alert_categories_per_month[col] = alert_categories_per_month[col].astype("int")

    return alert_categories_per_month



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

__version__ = "0.2.2"
description = "A lib for measuring ACE based IDR operations."

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Load requirements
with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    requirements = [line.strip() for line in f.readlines()]

setup(
    name="ace-metrics",
    version=__version__,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seanmcfeely/ace-metrics",
    author="Sean McFeely",
    author_email="mcfeelynaes@gmail.com",
    license="GNU General Public License v3.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.6",
    keywords="Information Security,ACE,ACE Ecosystem",
    packages=find_packages(),
    package_data={
      'ace_metrics.plotly_dash': ['assets/*'],
    },
    include_package_data=True,
    install_requires=requirements,
    scripts=["ace-metrics"],
)



"""Utility functions for database operations and plotting."""

import sqlite3
from typing import Dict, List, Tuple, Union

import pandas as pd

from .database import DATA_DIR

VALID_TABLE_NAMES = [
    "analyst_alert_quantities",
    "alert_type_quantities",
    "hours_of_operation",
    "overall_operating_alert",
    "cycle_time_sum",
    "cycle_time_mean",
    "cycle_time_min",
    "cycle_time_max",
    "cycle_time_std",
    "alert_count",
    "cycle_time_sum_BH",
    "cycle_time_mean_BH",
    "cycle_time_min_BH",
    "cycle_time_max_BH",
    "cycle_time_std_BH",
]


def fetch_data(tables: List[str], clean_column_names: bool = False) -> Dict[str, pd.DataFrame]:
    """Fetch tables from SQLite database.

    Args:
        tables: A list of table names to fetch data from.
        clean_column_names: Flag to clean column names (rename 'index'
          to 'month' and lowercase all column names).

    Returns:
        A dictionary where keys are table names and values are
          DataFrames with the fetched data.

    Raises:
        ValueError: If any of the table names provided are not valid.
    """
    conn = sqlite3.connect(f"{DATA_DIR}/ace_metrics_database.sqlite")
    df_dict = {}
    for table in tables:
        if table not in VALID_TABLE_NAMES:
            raise ValueError("Invalid table name")
        df = pd.read_sql_query(f"SELECT * FROM {table};", conn)
        if clean_column_names:
            df.rename(columns={"index": "month"}, inplace=True)
            df.columns = df.columns.str.lower()
        df_dict[f"{table}"] = df
    conn.close()
    return df_dict


def to_ordinal(n: int) -> str:
    """Convert an integer to its ordinal representation as a string."""
    suffix = ["th", "st", "nd", "rd", "th"][min(n % 10, 4)]
    if 11 <= (n % 100) <= 13:
        suffix = "th"
    return str(n) + suffix


def to_fiscal_year(year_month_str: str) -> str:
    """Convert a `YYYYMM` string to its corresponding fiscal year."""
    year = int(year_month_str[:4])
    month = int(year_month_str[4:])
    if month >= 10:
        year += 1
    return str(year)


def double_click_reset_y_range(
    fig_data: List[Dict], fig_layout: Dict, side_by_side_plot: bool = False
) -> Union[Tuple[float, float], List[Tuple[float, float]]]:
    """Reset the y-range when double-clicking plot area.

    Args:
        fig_data: A list of dicts, each representing a trace of data.
        fig_layout: A dict with the layout configuration of the figure.
        side_by_side_plot: Indicates if the plot is a side-by-side plot.

    Returns:
        A tuple containing the new y-axis range for the plot. If the
          plot is side by side, returns a list of tuples with the range
          for each plot.
    """
    x_axis = fig_data[0]["x"]
    cur_xrange = fig_layout["xaxis"]["range"]
    y1_min, y1_max = float("inf"), float("-inf")
    if side_by_side_plot:
        cur_xrange2 = fig_layout["xaxis2"]["range"]
        y2_min, y2_max = float("inf"), float("-inf")
    # Loop through all traces to find the min and max y-values within
    # the current x-axis range of each subplot
    for trace in fig_data:
        if trace.get("visible") == "legendonly":
            continue  # Skip this trace if it's not visible

        x_values, y_values = trace["x"], trace["y"]

        for x, y in zip(x_values, y_values):
            # Since both subplots share the legends from the first plot,
            # the second subplot will have showlegend=False.
            if "showlegend" not in trace or trace.get("showlegend") is True:
                if (
                    x_axis[max(0, int(-(-cur_xrange[0] // 1)))]
                    <= x
                    <= x_axis[min(int(cur_xrange[1] // 1), len(x_axis) - 1)]
                ):
                    y1_min = min(y1_min, y)
                    y1_max = max(y1_max, y)
            elif side_by_side_plot:
                if (
                    x_axis[max(0, int(-(-cur_xrange2[0] // 1)))]
                    <= x
                    <= x_axis[min(int(cur_xrange2[1] // 1), len(x_axis) - 1)]
                ):
                    y2_min = min(y2_min, y)
                    y2_max = max(y2_max, y)
    if side_by_side_plot:
        return [(y1_min, y1_max), (y2_min, y2_max)]
    return (y1_min, y1_max)


def change_color_brightness(hex_color: str, factor: float) -> str:
    """Change the brightness of a color represented in hex format."""
    hex_color = hex_color.lstrip("#")

    if len(hex_color) == 3:
        hex_color = "".join([x * 2 for x in hex_color])
    elif len(hex_color) != 6:
        raise ValueError("Input hex color should be 3 or 6 characters long after removing '#'")

    try:
        r, g, b = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))
    except ValueError:
        raise ValueError("Invalid hex color")

    # Darken the color by the given factor
    r = int(r * factor)
    g = int(g * factor)
    b = int(b * factor)

    # Ensure that the values are within the 0-255 range
    r = max(0, min(255, r))
    g = max(0, min(255, g))
    b = max(0, min(255, b))

    # Convert the RGB color back to hex
    darker_hex_color = "#{:02x}{:02x}{:02x}".format(r, g, b)

    return darker_hex_color



"""Hours of operations page for Dash app.

All functions that generate charts will have a @callback decorator along
  with its input and output parameters. They also share the return type
  and an exception as described below.

Args:
    relayoutData: Contains layout changes such as hidden traces, or user
      interactions like zooming, resetting axes, etc.
    template_data: Plot template based on chosen theme and color mode.
    current_fig: Existing figure to update, if applicable.

Returns:
    An updated figure or the same figure if no changes are made.

Raises:
    PreventUpdate: If no update is required or possible.
"""

from typing import Dict, List, Union

import dash
import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from dash import Input, Output, State, callback, dcc

from ace_metrics.plotly_dash.app import chart_config
from ace_metrics.plotly_dash.helpers import double_click_reset_y_range, to_fiscal_year, to_ordinal, fetch_data

dash.register_page(__name__, path="/hours-of-operations")


def generate_yearly_alert_count_df(hours_of_operation_df: pd.DataFrame) -> pd.DataFrame:
    """Aggregate alert counts by operating hours per fiscal year."""
    df = hours_of_operation_df.copy(True).set_index("month")
    df = df[["bh_day_quantities", "nights_quantities", "weekend_quantities"]]

    total = df.sum()
    total.name = "total"
    last_12_months = df.tail(12).sum()
    last_12_months.name = "last_12_months"

    df = df.assign(fiscal_year=df.index.astype(str).to_series().apply(to_fiscal_year))
    df = df.groupby("fiscal_year").sum()

    df = df.append([last_12_months, total])
    return df


layout = dbc.Container(
    [
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(
                            id="hours-of-operation",
                            config=chart_config,
                            style={"height": 405},
                        ),
                    ],
                    width={"size": 12, "order": "last", "offset": 0},
                ),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(
                    [
                        dcc.Graph(
                            id="yearly-hours-of-operation",
                            config={**chart_config, **{"doubleClick": "autosize"}},
                            style={"height": 405},
                        ),
                    ],
                    width={"size": 6, "order": "last", "offset": 0},
                ),
                dbc.Col(
                    [
                        dcc.Dropdown(
                            id="year-dropdown",
                            options=[
                                {"label": year, "value": year}
                                for year in generate_yearly_alert_count_df(
                                    fetch_data(["hours_of_operation"], True)["hours_of_operation"]
                                ).index
                            ],
                            value=["last_12_months", "total"],
                            multi=True,
                        ),
                        dcc.Graph(
                            id="yearly-alert-counts-by-operating-hours",
                            config={**chart_config, **{"doubleClick": "autosize"}},
                            style={"height": 405},
                        ),
                    ],
                    width={"size": 6, "order": "last", "offset": 0},
                ),
            ]
        ),
    ]
)


@callback(
    Output("yearly-alert-counts-by-operating-hours", "figure"),
    [Input("yearly-alert-counts-by-operating-hours", "relayoutData"), Input("theme-template-store", "data")],
    [Input("year-dropdown", "value")],
    [State("yearly-alert-counts-by-operating-hours", "figure")],
)
def update_yearly_alert_counts_by_operating_hours(
    relayoutData: Dict, template_data: Dict[str, Dict], selected_years: List[str], current_fig: Dict
) -> Union[go.Figure, Dict]:
    """Update or create chart of yearly alert counts by operating hours.

    Refresh the pie chart based on the user's year selection from the
      dropdown or the chosen legends. Update the layout and annotations
      of the chart accordingly. If no years are selected or if there's
      an attempt to update without necessary parameters, prevent the
      update to avoid errors.

    Args:
        selected_years: Years chosen by the user to display.
    """
    # If the plot already exists
    if relayoutData is not None and "hiddenlabels" in relayoutData:
        # Empty "hiddenlabels" dict means that all legends are selected
        if not relayoutData["hiddenlabels"]:
            pass
        else:
            return current_fig  # Redraw the plot with disabled legends

    # If the plot doesn't exist, create one
    if (not current_fig and relayoutData) or selected_years:
        df = generate_yearly_alert_count_df(fetch_data(["hours_of_operation"], True)["hours_of_operation"])
        df = df.rename(
            columns={
                "bh_day_quantities": "Business Hours",
                "nights_quantities": "Nights",
                "weekend_quantities": "Weekend",
            }
        )

        fig = go.Figure()
        if selected_years:
            width_per_pie = 1.0 / len(selected_years)
        else:
            return fig

        for i, year in enumerate(selected_years):
            values = df.loc[year, :]
            x_start = i * width_per_pie
            x_end = x_start + width_per_pie
            fig.add_trace(
                go.Pie(
                    labels=values.index,
                    values=values,
                    name=year,
                    textinfo="percent+value",
                    marker=dict(line=dict(color="white", width=2)),
                    textposition="inside",
                    insidetextorientation="tangential",
                    textfont=dict(size=15),
                    domain=dict(x=[x_start, x_end], y=[0, 1]),  # Set the domain for this pie chart
                )
            )
            fig.add_annotation(
                x=(x_start + x_end) / 2,
                y=-0.1,
                text=year,
                xanchor="center",
                showarrow=False,
                font=dict(size=16),
            )

        fig.update_layout(
            title="Yearly Alert Counts by Operating Hours",
            margin=dict(t=43, l=0, r=0),
            legend=dict(
                x=0,
                y=-0.1,
                traceorder="normal",
                orientation="h",
                title=dict(text=""),
            ),
            template=template_data["template"],
        )
        return fig

    raise dash.exceptions.PreventUpdate


@callback(
    Output("yearly-hours-of-operation", "figure"),
    [Input("yearly-hours-of-operation", "relayoutData"), Input("theme-template-store", "data")],
    [State("yearly-hours-of-operation", "figure")],
)
def update_yearly_hours_of_operation(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> go.Figure:
    """Update or create the yearly hours of operation bar chart.

    Generate a bar chart to represent average operation times across
      different periods, grouped by fiscal year and overall averages.
      Process, rank, and merge data before plotting. Show charts for the
      last 12 months and total averages by default.
    """
    if not current_fig and relayoutData:
        df = fetch_data(["hours_of_operation"], True)["hours_of_operation"].set_index("month")
        df = df[["bh_day_cycle_time_averages", "nights_cycle_time_averages", "weekend_cycle_time_averages"]]
        total = df.mean()
        total.name = "total"
        last_12_months = df.tail(12).mean()
        last_12_months.name = "last_12_months"

        df = df.assign(fiscal_year=df.index.astype(str).to_series().apply(to_fiscal_year))
        df = df.groupby("fiscal_year").mean()

        df = df.append([last_12_months, total]).round(1)
        df.rename(
            columns={
                "bh_day_cycle_time_averages": "Business Hours",
                "nights_cycle_time_averages": "Nights",
                "weekend_cycle_time_averages": "Weekend",
            },
            inplace=True,
        )
        df.index.name = "index"
        df = df.T

        sorted_df = pd.DataFrame()
        for year_time in df.columns:
            temp_df = df.sort_values(by=year_time, ascending=False).reset_index()
            temp_df["rank"] = temp_df.index.map(lambda x: f"{to_ordinal(x+1)}")
            temp_df = temp_df.rename(columns={"index": f"averages_{year_time}"})
            temp_df = temp_df[["rank", f"averages_{year_time}", year_time]]

            if sorted_df.empty:
                sorted_df = temp_df
            else:
                sorted_df = pd.merge(sorted_df, temp_df, on="rank", how="outer")

        # Create the figure
        fig = go.Figure()
        for year_time in df.columns:
            fig.add_trace(
                go.Bar(
                    x=sorted_df["rank"],
                    y=sorted_df[year_time],
                    name=f"{year_time}",
                    text=sorted_df[f"averages_{year_time}"],
                    hoverinfo="y+text",
                )
            )
        fig.update_layout(
            title="Yearly Hours of Operation Rankings",
            xaxis_title="rank",
            yaxis_title="hours",
            margin=dict(t=43, l=0, r=0),
            template=template_data["template"],
        )
        for trace in fig.data:
            if trace.name in ["last_12_months", "total"]:
                trace.visible = True
            else:
                trace.visible = "legendonly"

        return fig

    raise dash.exceptions.PreventUpdate


@callback(
    Output("hours-of-operation", "figure"),
    [Input("hours-of-operation", "relayoutData"), Input("theme-template-store", "data")],
    [State("hours-of-operation", "figure")],
)
def update_hours_of_operation(
    relayoutData: Dict, template_data: Dict[str, Dict], current_fig: Dict
) -> Union[go.Figure, Dict]:
    """Update or create the monthly hours of operation line chart.

    Generate a line chart if it doesn't exist. If the chart already
      exists, double-clicking on it will reset the range of y-axis.
    """
    if not current_fig and relayoutData:
        df = fetch_data(["hours_of_operation"], True)["hours_of_operation"].set_index("month")
        df = df[["bh_day_cycle_time_averages", "nights_cycle_time_averages", "weekend_cycle_time_averages"]]
        fig = px.line(
            df,
            x=df.index,
            y=df.columns.tolist(),
            labels={"value": ""},
            title="Monthly Hours of Operation",
            range_x=[len(df.index) - 25, len(df.index) - 1],
            range_y=[0, df.tail(25).max().max()],
        )

        fig.update_layout(
            legend={"title_text": ""},
            yaxis_title="hours",
            margin=dict(t=43, l=0, r=0),
            template=template_data["template"],
        )
        for trace in fig.data:
            if trace.name == "bh_day_cycle_time_averages":
                trace.name = "Business Hours"
            elif trace.name == "nights_cycle_time_averages":
                trace.name = "Nights"
            else:
                trace.name = "Weekend"
            trace.mode = "lines+markers"
            trace.marker = dict(size=5, color=trace.line.color)

        return fig

    if current_fig and relayoutData == {}:
        current_fig["layout"]["yaxis"]["range"] = double_click_reset_y_range(current_fig["data"], current_fig["layout"])
        return current_fig

    raise dash.exceptions.PreventUpdate
